/* normalmode_copy.cc
 * This file belongs to Worker, a file manager for UN*X/X11.
 * Copyright (C) 2011-2013 Ralf Hoffmann.
 * You can contact me at: ralf@boomerangsworld.de
 *   or http://www.boomerangsworld.de/worker
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "normalmode.h"
#include "worker_locale.h"
#include "worker.h"
#include "copyopwin.hh"
#include "copyop.h"
#include "nmcacheentry.hh"
#include "nmcopyopdir.hh"
#include "nmfieldlistviewdnd.hh"
#include "datei.h"
#include "fileentry.hh"
#include "simplelist.hh"
#include "copycore.hh"
#include "copyorder.hh"
#include "virtualdirmode.hh"

void NormalMode::copy( struct copyorder *copyorder )
{
    Lister *olister=NULL;
    NormalMode *nm2=NULL,*updatenm=NULL;
    ListerMode *lm=NULL;
    NM_specialsourceInt *ss1;
    bool cancel=false;
    CopyOpWin *cowin;
    int erg;
    char *textstr, *buttonstr;
    std::list<NM_specialsourceInt*>::iterator iti1;
    std::list<NM_specialsourceInt*> *copy_lvc_list;
    std::list<const FileEntry*>::iterator itfe1;
    VirtualDirMode *updatevdm = NULL;
  
    if(copyorder==NULL) return;
    if(copyorder->destdir==NULL) return;
    if ( getCurrentDir() == NULL ) {
        // no current dir => no files to copy
        return;
    }

    if ( ce->dirOpened() == false ) return;

    copyorder->overwrite=copyorder->COPY_OVERWRITE_NORMAL;  // force to start in this mode

    finishsearchmode();
  
    if(Datei::fileExistsExt(copyorder->destdir)!=Datei::D_FE_DIR) {
        // no dir as destination
        textstr=(char*)_allocsafe(strlen(catalog.getLocale(67))+strlen(copyorder->destdir)+1);
        sprintf(textstr,catalog.getLocale(67),copyorder->destdir);
        buttonstr = dupstring( catalog.getLocale(11) );
        request(catalog.getLocale(125),textstr,buttonstr);
        _freesafe(textstr);
        _freesafe( buttonstr );
        return;
    }
  
    disableDirWatch();
  
    // clear the reclist so the slace doesn't block us because he is still reading
    // from disk
    ft_request_list_clear();
  
    if ( strcmp( getCurrentDir(), copyorder->destdir ) == 0 ) {
        copyorder->do_rename=true;
        updatenm=this;
    } else {
        olister=parentlister->getWorker()->getOtherLister(parentlister);
        if(olister!=NULL) {
            lm=olister->getActiveMode();
            if(lm!=NULL) {
                if(lm->isType("NormalMode")==true) {
                    nm2=(NormalMode*)lm;
                    if ( nm2->getCurrentDir() != NULL ) {
                        if ( strcmp( nm2->getCurrentDir(), copyorder->destdir ) == 0 ) {
                            updatenm=nm2;
                        }
                    }
                } else if ( ( updatevdm = dynamic_cast< VirtualDirMode * >( lm ) ) ) {
                    if ( updatevdm->getCurrentDirectory() != copyorder->destdir ) {
                        updatevdm = NULL;
                    }
                }
            }
        }
    }
    if ( updatenm != NULL && updatenm != this ) {
        updatenm->disableDirWatch();
    } else if ( updatevdm != NULL ) {
        updatevdm->disableDirWatch();
    }
  
    // no sense for follow_symlinks when moving
    if(copyorder->move==true) copyorder->follow_symlinks=false;

    /* now presupposition checked
       next search for the files to copy
       therefore collect entries to copy in a list together with the LVC (when on root-level)
       then for each entry do recursive call to copy it
       after success deactivate LVC */

    /* first: create List contains NM_specialsource
       when source==COPY_SPECIAL then the given list but check for existing of the FEs
       else source==COPY_ONLYACTIVE => take the activefe (from ce) in this list
       else take all selected entries (except the "..")
    */
    cowin=copyorder->cowin;
    if(cowin!=NULL) {
        cowin->open();
        cowin->setmessage(catalog.getLocale(122),0);
        if ( cowin->redraw() != 0 ) cancel = true;
    }

    CopyCore *cc = new CopyCore( copyorder );
    int row_adjustment = 0;

    if ( cancel ) cc->setCancel( cancel );

    {
        std::list< NM_specialsourceInt *> copylist;
        switch(copyorder->source) {
            case copyorder::COPY_SPECIAL:
                buildSpecialSourceIntList( *( copyorder->sources ),
                                           copylist );
                break;
            case copyorder::COPY_ONLYACTIVE:
                getSelFiles( &copylist, LM_GETFILES_ONLYACTIVE, false );
                break;
            default:  // all selected entries
                getSelFiles( &copylist, LM_GETFILES_SELORACT, false );
                break;
        }
        for ( auto &it : copylist ) {
            cc->registerFEToCopy( *it->entry(), it->row );
        }

        for ( iti1 = copylist.begin(); iti1 != copylist.end(); iti1++ ) {
            ss1 = *iti1;
            if(ss1->cod!=NULL) delete ss1->cod;
            delete ss1;
        }
    }

    if (cc->buildCopyDatabase() != 0) cc->setCancel( true );

    // this is just the value from the last update, it might not be correct
    if ( parentlister->getWorker()->PS_readSpace( copyorder->destdir ) == 0 ) {
        loff_t free_blocks = parentlister->getWorker()->PS_getFreeSpace();
        loff_t block_size = parentlister->getWorker()->PS_getBlocksize();

        if ( block_size > 0 && ( ( cc->getBytesToCopy() ) / block_size ) + 1 >= free_blocks ) {
            // files will possibly not fit on target, ask to continue or cancel

            std::string buttons = catalog.getLocale( 629 );
            buttons += "|";
            buttons += catalog.getLocale( 8 );

            std::string freespaceh = parentlister->getWorker()->PS_getFreeSpaceH();

            std::string text = AGUIXUtils::formatStringToString( catalog.getLocale( 992 ),
                                                                 copyorder->destdir,
                                                                 freespaceh.c_str() );

            erg = request( cowin, catalog.getLocale( 123 ), text.c_str(), buttons.c_str() );

            if ( erg != 0 ) {
                cc->setCancel( true );
            }
        }
    }

    // now database ready, can start copy-process

    copy_lvc_list = new std::list<NM_specialsourceInt*>;

    cc->setPreCopyCallback( [this,
                             &row_adjustment]( const FileEntry &fe, int row, const NM_CopyOp_Dir *cod )
                            {
                                int actual_row = row - row_adjustment;
                                if ( actual_row >= 0 ) {
                                    lv->setVisMark( actual_row, true );
                                    lv->showRow( actual_row, false );

                                    // this is some optimization when copying a lot of small files
                                    // limit redrawing in this case
                                    if ( cod == NULL ) {
                                        loff_t size = 0;

                                        if ( fe.isLink && ! fe.isCorrupt ) {
                                            size = fe.dsize();
                                        } else {
                                            size = fe.size();
                                        }

                                        if ( size > 128 * 1024 ) {
                                            lv->redraw();
                                        } else {
                                            lv->timeLimitedRedraw();
                                        }
                                    } else {
                                        lv->redraw();
                                    }
                                }
                            } );

    cc->setPostCopyCallback( [this,
                              &copyorder,
                              &copy_lvc_list,
                              &cc,
                              &row_adjustment]( const FileEntry &fe, int row,
                                                CopyCore::nm_copy_t res, const NM_CopyOp_Dir *cod,
                                                const std::string &target_fullname )
                             {
                                 if ( copyorder->move == false ) {
                                     // just copy

                                     if ( ! cc->getCancel() &&
                                          ( res == CopyCore::NM_COPY_OK || res == CopyCore::NM_COPY_NEED_DELETE ) &&
                                          ( cod == NULL || cod->error_counter == 0 ) ) {
                                         // success

                                         FileEntry *real_fe = getFEForRow( row );

                                         if ( real_fe && real_fe->equals( fe ) ) {
                                             if ( row >= 0 ) {
                                                 lv->setSelect( row, false );
                                                 lv->showRow( row, false );
                                                 lv->setVisMark( row, false );
                                             }
                                             real_fe->select = false;
                                             ce->recalcStats();
                                             showCacheState();
                                         }
                                     } else {
                                         // failed

                                         if ( row >= 0 ) {
                                             lv->showRow( row, false );
                                             lv->setVisMark( row, false );
                                         }
                                     }
                                 } else {
                                     // move so remove entry if successful
                                     int actual_row = row - row_adjustment;

                                     FileEntry *real_fe = getFEForRow( actual_row );

                                     if ( real_fe && real_fe->equals( fe ) ) {

                                         lv->setVisMark( actual_row, false );

                                         if ( res == CopyCore::NM_COPY_OK &&
                                              ( cod == NULL || cod->error_counter == 0 ) ) {
                                             // success so remove entry
                                             removeEntry( real_fe, actual_row );

                                             row_adjustment++;
                                         } else if ( res == CopyCore::NM_COPY_NEED_DELETE ) {
                                             if ( actual_row >= 0 ) {
                                                 NM_specialsourceInt *tss;

                                                 tss = new NM_specialsourceInt( real_fe );
                                                 tss->row = actual_row;
                                                 copy_lvc_list->push_back( tss );
                                             }
                                         } else if ( res == CopyCore::NM_COPY_CANCEL ) {
                                             cc->setCancel( true );
                                         }
                                     }
                                 }
                             } );

    if ( ! cc->getCancel() ) {
        cc->executeCopy();
    }

    lv->redraw();
  
    // now delete remaining files/dirs when moving
    if ( copyorder->move == true ) {

        auto iti2 = copy_lvc_list->begin();

        cc->deleteCopiedButNotMoved( [&iti2,
                                      &copy_lvc_list,
                                      this]( const std::pair< std::string, bool > &removed_entry )
                                     {
                                         NM_specialsourceInt *tss = NULL;

                                         if ( iti2 != copy_lvc_list->end() ) {
                                             tss = *iti2;
                                         }

                                         if ( tss != NULL ) {
                                             if ( removed_entry.first == tss->entry()->fullname ) {
                                                 lv->showRow( tss->row );
                                                 removeEntry( tss->entry(), tss->row );
                                                 fixSpecialSourceList( copy_lvc_list, tss->row );
                                                 iti2++;
                                             }
                                         }
                                     } );
    }

    delete cc;
    cc = NULL;
  
    if(cowin!=NULL) cowin->close();
  
    for ( iti1 = copy_lvc_list->begin(); iti1 != copy_lvc_list->end(); iti1++ ) {
        auto tss = *iti1;
        delete tss;
    } 

    delete copy_lvc_list;
    if(copyorder->move==true) {
        ce->restartcheck();
        rebuildView();
    }
    if(updatenm!=NULL) updatenm->update(false);
    else if ( updatevdm != NULL ) updatevdm->update();

    lv->showActive();
    aguix->Flush();

    enableDirWatch();
    if ( updatenm != NULL && updatenm != this ) {
        updatenm->enableDirWatch();
    } else if ( updatevdm != NULL ) {
        updatevdm->enableDirWatch();
    }
}
